import std/[json, jsonutils, logging, options, strutils, sequtils]
import jester, namespaced_logging

import ./apierror

const CONTENT_TYPE_JSON* = "application/json"

var logNs {.threadvar.}: LoggingNamespace

template log(): untyped =
  if logNs.isNil: logNs = getLoggerForNamespace("buffoonery/apiutils", lvlDebug)
  logNs

## Response Utilities
## ------------------

type ApiResponse*[T] = object
  details*: Option[string]
  data*: Option[T]
  nextOffset*: Option[int]
  totalItems*: Option[int]
  nextLink*: Option[string]
  prevLink*: Option[string]

func initApiResponse*[T](
    details = none[string](),
    data: Option[T] = none[T](),
    nextOffset = none[int](),
    totalItems = none[int](),
    nextLink = none[string](),
    prevLink = none[string]()): ApiResponse[T] =
  ApiResponse[T](details: details, data: data, nextOffset: nextOffset,
    totalItems: totalItems, nextLink: nextLink, prevLink: prevLink)

func `%`*(r: ApiResponse): JsonNode =
  result = newJObject()
  if r.details.isSome: result["details"] = %r.details
  if r.data.isSome: result["data"] = %r.data
  if r.nextOffset.isSome: result["nextOffset"] = %r.nextOffset
  if r.totalItems.isSome: result["totalItems"] = %r.totalItems
  if r.nextLink.isSome: result["nextLink"] = %r.nextLink
  if r.prevLink.isSome: result["prevLink"] = %r.prevLink

template halt*(
    code: HttpCode,
    headers: RawHeaders,
    content: string) =
  ## Immediately replies with the specified request. This means any further
  ## code will not be executed after calling this template in the current
  ## route.
  bind TCActionSend, newHttpHeaders
  result[0] = CallbackAction.TCActionSend
  result[1] = code
  result[2] = if isSome(result[2]): some(result[2].get & headers)
              else: some(headers)
  result[3] = content
  result.matched = true
  break allRoutes

template sendJsonResp*(
    body: JsonNode,
    code: HttpCode = Http200,
    knownOrigins: seq[string] = @[],
    headersToSend: RawHeaders = @{:}) =
  ## Immediately send a JSON response and stop processing the request.
  let reqOrigin =
    if headers(request).hasKey("Origin"): $(headers(request)["Origin"])
    else: ""

  let corsHeaders =
    if knownOrigins.contains(reqOrigin):
      @{
        "Access-Control-Allow-Origin": reqOrigin,
        "Access-Control-Allow-Credentials": "true",
        "Access-Control-Allow-Methods": $(reqMethod(request)),
        "Access-Control-Allow-Headers": "Authorization,X-CSRF-TOKEN"
      }
    else:
      log().debug "Unrecognized Origin '" & reqOrigin & "', excluding CORS headers."
      @{:}

  halt(
    code,
    cast[RawHeaders](headersToSend) & corsHeaders & @{
      "Content-Type": CONTENT_TYPE_JSON,
      "Cache-Control": "no-cache"
    },
    $body
  )

template sendResp*[T](
    resp: ApiResponse[T],
    code = Http200,
    allowedOrigins = newSeq[string](),
    headersToSend: RawHeaders = @{:}) =
  sendJsonResp(%resp, code, allowedOrigins, headersToSend)

template sendErrorResp*(err: ref ApiError, knownOrigins: seq[string]): void =
  log().error err.respMsg & ( if err.msg.len > 0: ": " & err.msg else: "")
  if not err.parent.isNil: log().error "    original exception: " & err.parent.msg
  sendJsonResp( %*{"details":err.respMsg}, err.respCode, knownOrigins)

## CORS support
template sendOptionsResp*(
    allowedMethods: seq[HttpMethod],
    knownOrigins: seq[string]) =

  let reqOrigin =
    if headers(request).hasKey("Origin"): $(headers(request)["Origin"])
    else: ""

  let corsHeaders =
    if knownOrigins.contains(reqOrigin):
      @{
        "Access-Control-Allow-Origin": reqOrigin,
        "Access-Control-Allow-Credentials": "true",
        "Access-Control-Allow-Methods": allowedMethods.mapIt($it).join(", "),
        "Access-Control-Allow-Headers": "DNT,User-Agent,X-Requested-With,If-Modified-Since,Cache-Control,Content-Type,Range,Authorization,X-CSRF-TOKEN"
      }
    else:
      log().debug "Unrecognized Origin '" & reqOrigin & "', excluding CORS headers."
      log().debug "Valid origins: " & knownOrigins.join(", ")
      @{:}

  halt(
    Http200,
    corsHeaders,
    ""
  )
